----------------------------------------------------------------------------------------------------------------------------------
-- QuickTalents
----------------------------------------------------------------------------------------------------------------------------------
CreateFrame("Frame","QuickTalents",UIParent):RegisterEvent("ADDON_LOADED");
QuickTalents:SetScript("OnEvent",function(self)
	
	-- load settings
	local settings = {
		ButtonSize = {"number",28,"Size of the Icons."},
		ShowReagents = {"boolean",true,"Whether or not to display Reagent count."},
		ShowTooltips = {"boolean",false,"Whether or not to show tooltips over icons."},
		BackgroundAlpha = {"number",0.75,"Visibility of the frame background."},
		ShowGlyphs = {"boolean",true,"Enable/Disable glyph buttons."},
		GlyphHistorySize = {"number",3,"Number of previously used glyphs displayed."},
		Position = {"table",{"TOPRIGHT",-5,-210}},
		Bindings = {"table",{}}, -- TODO: maybe make this character specific.
		GlyphHistory = {"table",{}},
	};
	QT_Saved = QT_Saved or {};
	local cfg = QT_Saved;
	for k,v in pairs(settings) do
		cfg[k] = ( cfg[k] ~= nil and type(cfg[k]) == v[1] ) and cfg[k] or v[2];
	end
	
	-- Learn Queue
	local Queue = {};
	function self:Learn(t)
		PlayerTalentFrame_ClearTalentSelections();
		local r = ceil(t/3);
		Queue[r] = nil;
		for i = r*3-2, r*3 do
			if select(5,GetTalentInfo(i)) then
				Queue[r] = t;
				return
			end
		end
		LearnTalents(t);
	end
	
	-- Glyphs
	local GlyphHistory,PlayerSpec,PlayerGlyphs;
	function self:UpdateGlyphs()
		if PlayerSpec ~= GetSpecialization() then
			PlayerSpec = GetSpecialization()
			-- get players current glyphs
			PlayerGlyphs = wipe(PlayerGlyphs or {});
			for i = 1,3 do
				local spell,icon,id = select(4,GetGlyphSocketInfo(i*2));
				PlayerGlyphs[i] = id and { id, GetSpellInfo(spell):sub(10), icon } or {};
			end
			-- load glyph history
			local class = select(2,UnitClass("Player"));
			cfg.GlyphHistory[class] = cfg.GlyphHistory[class] or {};
			if PlayerSpec then
				cfg.GlyphHistory[class][PlayerSpec] = cfg.GlyphHistory[class][PlayerSpec] or {};
			end
			GlyphHistory = cfg.GlyphHistory[class][PlayerSpec];
		end
		if not (PlayerGlyphs and GetGlyphSocketInfo(1) and GlyphHistory) then return end
		local h = GlyphHistory;
		for i = 1,3 do
			local spell,icon,id = select(4,GetGlyphSocketInfo(i*2));
			if PlayerGlyphs[i][1] ~= id then -- glyph slot has changed
				if id then -- remove new glyph from history
					local found = 0;
					for j = 1,#h do
						if h[j][1] == id then found = 1; end
						h[j] = h[j+found];
					end
				end
				if PlayerGlyphs[i][1] then -- add previous glyph to history
					for j = #h,0,-1 do
						h[j+1] = h[j] or PlayerGlyphs[i];
					end
				end
				PlayerGlyphs[i] = id and { id, GetSpellInfo(spell):sub(10), icon } or {}; -- update PlayerGlyphs
			end
		end
	end
	
	-- Event Handler
	self:RegisterEvent("PLAYER_TALENT_UPDATE");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("BAG_UPDATE_DELAYED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:SetScript("OnEvent",function(self,e,...)
		for i,t in pairs(Queue) do self:Learn(t); end
		self:UpdateGlyphs();
		self:UpdateFrames(e);
	end);
	
	-- Setup main frame
	self:SetPoint(unpack(cfg.Position));
	self:CreateTexture("QuickTalentsBackground"):SetAllPoints();
	self:CreateFontString("QuickTalentsReagents"):SetFont("Fonts\\ARIALN.TTF", 13, "OUTLINE");
	self:EnableMouse(true);
	self:SetMovable(true);
	self:RegisterForDrag("LeftButton")
	self:SetScript("OnDragStart", self.StartMoving)
	self:SetScript("OnDragStop", function()
		self:StopMovingOrSizing();
		cfg.Position = {self:GetPoint()};
	end)
	
	-- Frame Handler
	local frames = nil;
	function self:SetupFrames()
		frames = frames or {};
		for i = 1,21+cfg.GlyphHistorySize do
			local col,row = (i-1)%3,floor((i-1)/3);
			
			frames[i] = frames[i] or CreateFrame("BUTTON","QuickTalentsButton"..i,self,"SecureActionButtonTemplate");
			frames[i]:SetAttribute("type1","macro");
			frames[i]:SetSize(cfg.ButtonSize-3,cfg.ButtonSize-3);
			frames[i].t = frames[i].t or frames[i]:CreateTexture("QuickTalentsButton"..i.."Texture");
			frames[i].t:SetAllPoints();
			frames[i].t:SetTexCoord(0.075,0.925,0.075,0.925);
			frames[i]:SetScript("OnLeave",function() GameTooltip:Hide(); end);
			frames[i]:SetScript("OnEnter",self.ButtonHover);
			local offset = i <= 18 and 2 or ( cfg.ShowReagents and 16 or 6 );
			frames[i]:SetPoint("TOPLEFT",col*cfg.ButtonSize+2,-(row*cfg.ButtonSize+offset));
			
			if i <= 18 then -- talents
				frames[i]:SetAttribute("macrotext",
					"/stopmacro [combat]\n"..
					"/click QuickTalentsOpener\n".. -- ensures the talent frame is ready for interaction
					"/click PlayerTalentFrameTalentsTalentRow"..(row+1).."Talent"..(col+1).."\n".. -- only way(?) to get the unlearn popup without taint
					"/click StaticPopup1Button1\n".. -- confirm unlearn (TODO: what if popup1 is not the talent prompt)
					"/run QuickTalents:Learn("..i..")\n" -- queue new talents for learn
				);
				frames[i]:RegisterForDrag("LeftButton");
				frames[i]:SetScript("OnDragStart",function()
					if not InCombatLockdown() then
						PickupTalent(i)
						if CursorHasSpell() then
							QuickTalentsBinder.spell = select(4,GetCursorInfo());
							QuickTalentsBinder:SetScript("OnUpdate",QuickTalentsBinder.OnEvent);
						end
					end
				end);
			elseif i <= 21 then -- glyphs slots
				frames[i]:SetAttribute("macrotext",format("/click GlyphFrameGlyph%d\n/click StaticPopup1Button1\n",(i-18)*2));
			else -- glyph history
				frames[i].t:SetVertexColor(1,1,1,0.25);
				frames[i]:RegisterForClicks("RightButtonUp","LeftButtonDown");
				frames[i]:SetAttribute("type2","script"); -- TODO: maybe use a modified click instead
				frames[i]:SetAttribute("_script",function(btn) -- remove from history
					for j = i-21,#GlyphHistory do GlyphHistory[j] = GlyphHistory[j+1]; end
					self:UpdateFrames();
				end);
			end
			if not cfg.ShowGlyphs and i > 18 then frames[i]:Hide(); end
		end
	end
	function self:UpdateFrames(e)
		for i = 1,18 do -- talent textures
			frames[i].t:SetVertexColor(1,1,1,select(5,GetTalentInfo(i)) and 1 or 0.25);
			frames[i].t:SetTexture((select(2,GetTalentInfo(i))));
		end
		if e and e:sub(1,12) == "PLAYER_REGEN" then -- combat fading
			for i = 1,21+cfg.GlyphHistorySize do
				SetDesaturation(frames[i].t, e:sub(14) == "DISABLED");
			end
		end
		if cfg.ShowReagents then QuickTalentsReagents:SetText(select(2,GetTalentClearInfo())); end -- reagent count
		if cfg.ShowGlyphs and GlyphHistory and e ~= "PLAYER_REGEN_DISABLED" and not InCombatLockdown() then
			for i = 19,21+cfg.GlyphHistorySize do
				local icon,id,name;
				if i <= 21 then -- glyph slots
					icon,id = select(5,GetGlyphSocketInfo((i-18)*2));
				elseif GlyphHistory[i-21] then -- glyph history
					id,name,icon = unpack(GlyphHistory[i-21]);
					frames[i]:SetAttribute("macrotext",
						"/stopmacro [combat]\n"..
						"/click QuickTalentsOpener\n"..
						"/run SetGlyphNameFilter(\""..name.."\")\n".. -- set name filter
						"/run if IsGlyphFlagSet(1) then ToggleGlyphFilter(1) end\n".. -- prep header
						"/click GlyphFrameHeader1\n".. -- trigger scrollframe update
						"/click GlyphFrameScrollFrameButton2\n" -- click glyph button, TODO: are there glyphs that return multiple results?
					);
				else
					frames[i]:SetAttribute("macrotext", "");
				end
				frames[i].t:SetTexture(icon or 0.5,0.5,0.5,0.5);
				frames[i].glyphID = id;
				frames[i]:Show();
			end
		end
	end
	function self.ButtonHover(btn)
		if cfg.ShowTooltips then
			GameTooltip:SetOwner(btn, "ANCHOR_RIGHT");
			if btn.glyphID then GameTooltip:SetGlyphByID(btn.glyphID);
			else GameTooltip:SetTalent(btn:GetName():sub(19)); end
			GameTooltip:Show();
		end
	end
	
	-- Handles the safe loading & opening of the Blizzard Talent UI
	CreateFrame("BUTTON","QuickTalentsOpener",self,"SecureActionButtonTemplate"):SetAttribute("type","macro");
	QuickTalentsOpener:SetAttribute("macrotext",
		"/click TalentMicroButton\n"..
		"/click [spec:1] PlayerSpecTab1\n"..
		"/click [spec:2] PlayerSpecTab2\n"..
		"/click PlayerTalentFrameTab3\n"..
		"/click PlayerTalentFrameTab2\n"..
		"/click TalentMicroButton"
	);
	
	-- Binder
	CreateFrame("Frame","QuickTalentsBinder",self);
	function QuickTalentsBinder:OnEvent(e,arg1)
		if e == "LEARNED_SPELL_IN_TAB" then
			if cfg.Bindings[arg1] then
				PickupSpell(arg1);
				PlaceAction(cfg.Bindings[arg1]);
				ClearCursor();
			end
		elseif self.spell and select(4,GetCursorInfo()) ~= self.spell then
			if e == "ACTIONBAR_SLOT_CHANGED" then
				PlaySound("TellMessage");
				print(format("|HQT:bind:%d:%d|h[Click Here]|h",self.spell,arg1).." if you want to bind "..GetSpellLink(self.spell).." to action slot "..arg1..".");
			end
			self.spell = nil;
			self:SetScript("OnUpdate",nil);
		end
	end
	QuickTalentsBinder:RegisterEvent("ACTIONBAR_SLOT_CHANGED");
	QuickTalentsBinder:RegisterEvent("LEARNED_SPELL_IN_TAB");
	QuickTalentsBinder:SetScript("OnEvent",QuickTalentsBinder.OnEvent)
	
	-- hook chatframe hyperlink handler for bind links
	local _OHLC = ChatFrame1:GetScript("OnHyperLinkClick");
	ChatFrame1:SetScript("OnHyperLinkClick",function(self,link,text,btn)
		local arg = {strsplit(":",link)};
		if arg[1] == "QT" then
			if arg[2] == "bind" then
				cfg.Bindings[tonumber(arg[3])] = tonumber(arg[4]);
				print(GetSpellLink(arg[3]).." bound to action slot "..arg[4]..".");
			end
			return;
		end
		_OHLC(self,link,text,btn);
	end);
	
	-- Settings Editor
	SLASH_QUICKTALENTS1,SLASH_QUICKTALENTS2 = "/quicktalents","/qts";
	SlashCmdList.QUICKTALENTS = function(args)
		local arg1,arg2 = strsplit(" ",args);
		for k,v in pairs(settings) do
			if v[3] and arg2 and ( strlower(arg1) == strlower(k) or arg1 == k:gsub("[^A-Z]","") ) then
				if settings[k][1] == "boolean" then
					cfg[k] = strlower(arg2) == "true";
				elseif settings[k][1] == "number" then
					cfg[k] = tonumber(arg2);
				end
				self:SettingsChanged();
				print(k..": [|cFF3399FF"..tostring(cfg[k]).."|r]");
				return;
			end
		end
		print("|cFF009999QuickTalents Settings:|r (/qts [setting] [value])");
		for k,v in pairs(settings) do
			if v[3] then -- has description/is user setting
				print(k:gsub("[A-Z]","|cFFCCFFAA%1|r")..": [|cFF3399FF"..tostring(cfg[k]).."|r] - |cFFAAAAAA"..v[3].."|r");
			end
		end
	end
	
	-- Updates frames after setting changes
	function self:SettingsChanged()
		local rows = 6+(cfg.ShowGlyphs and ceil(cfg.GlyphHistorySize/3+1) or 0)
		self:SetSize( cfg.ButtonSize*3+1, cfg.ButtonSize*rows+(cfg.ShowReagents and 15 or 1) );
		self:SetupFrames();
		for i = 21+cfg.GlyphHistorySize,#frames do frames[i]:Hide(); end
		QuickTalentsBackground:SetTexture( 0,0,0, cfg.BackgroundAlpha );
		QuickTalentsReagents:SetText("");
		QuickTalentsReagents:SetPoint("TOPLEFT",2,-(cfg.ButtonSize*6));
		self:UpdateFrames();
	end
	self:SettingsChanged();
	
end);